/* ====================================================================
 * Copyright (c) 1996 NeoSoft, Inc.  All rights reserved.
 * Based on work Copyright (c) 1995 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the NeoSoft, Inc,
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server", "Apache Group", "NeoScript", "NeoWebScript",
 *    and "NeoSoft, Inc", must not be used to endorse or promote products 
 *    derived from this software without prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by NeoSoft, Inc, and
 *    the Apache Group for use in the Apache HTTP server project 
 *    (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY NEOSOFT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL NEOSOFT OR THE APACHE GROUP 
 * OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */


/*
 * mod_neoinclude.c: Handles NeoWebScript server-parsed HTML documents
 * 
 * Based on include processing module originally written by Rob McCool; 
 * with substantial fixups by David Robinson; 
 * incorporated into the Shambhala module framework by rst.
 * 
 */

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"

#include <db.h>
#include <assert.h>

#include "tcl.h"
#include "tclExtend.h"

module includes_and_neoscript_module;

extern char server_root[];
extern char server_confname[];

Tcl_Interp *interp;
void Tcl_InitExtensions (Tcl_Interp *interp);

char softwareStartTimeString[32];

#define NEOSCRIPT_VERSION "1.0"

#define STARTING_SEQUENCE "<!--#"
#define ENDING_SEQUENCE "-->"
#define DEFAULT_ERROR_MSG "[an error occurred while processing this directive]"
#define DEFAULT_TIME_FORMAT "%A, %d-%b-%y %T %Z"
#define SIZEFMT_BYTES 0
#define SIZEFMT_KMG 1

/* this will be relative to the server root, hopefully, or if you
 * set it to start with /, from there.
 */
#define URL_ACCESS_COUNTER_DB "neoscript-data/system/url_access_counters.db"

static void decodehtml(char *s);
static char *get_tag(pool *p, FILE *in, char *tag, int tag_len, int dodecode);
static int get_directive(FILE *in, char *d);

static request_rec *Tcl_request_rec = NULL;

extern int Neo_UnescapeStringCmd (ClientData clientData, Tcl_Interp *interp, int argc, char **argv);

module includes_and_neoscript_module;
enum xbithack { xbithack_off, xbithack_on, xbithack_full };

#ifdef XBITHACK	
#define DEFAULT_XBITHACK xbithack_full
#else
#define DEFAULT_XBITHACK xbithack_off
#endif

typedef struct {
    table *neowebscript_dir_vars;
    enum xbithack xbithack;
} neoscript_dir_config;

typedef struct {
    table *neowebscript_server_vars;
} neoscript_server_config;

int load_sub_req (Tcl_Interp *interp, request_rec *r)
{
    int errstatus; 
    int fd;
    char *buffer;
    int nLines;
    int whichLine;
 
    if (r->method_number != M_GET) return DECLINED;
    if (r->finfo.st_mode == 0 || (r->path_info && *r->path_info)) {
        log_reason("File does not exist", r->filename, r);
        return NOT_FOUND;
    } 
  
    if ((errstatus = set_content_length (r, r->finfo.st_size)) 
        || (errstatus = set_last_modified (r, r->finfo.st_mtime)))  
        return errstatus;
  
    fd = open (r->filename, O_RDONLY); 
 
    if (fd == -1) {
        log_reason("file permissions deny server access", r->filename, r);
        return FORBIDDEN;
    }

    buffer = (char *)ckalloc(r->finfo.st_size + 1);
    if (read(fd,buffer,r->finfo.st_size) != r->finfo.st_size) {
	if (errno != EINTR) goto bye;
    }
    buffer[r->finfo.st_size] = '\0';
    if (Tcl_GlobalEval(interp, buffer) != TCL_OK) {
	log_printf(r->server, "Error loading file '%s': %s", r->filename, interp->result);
	rprintf(r,"[error loading %s]", r->filename);
	close (fd);
	return SERVER_ERROR;
    }
    bye:
    close (fd);
    return OK;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Neo_IncludeCmd --
 *     Implements the Neo TCL web load and include commands:
 *         SAFE_include_file slave file
 *         SAFE_include_virtual slave file
 *         SAFE_load_file slave file
 *         SAFE_load_virtual slave file
 *
 * Results:
 *     Standard TCL results.
 *
 *-----------------------------------------------------------------------------
 */
int
Neo_IncludeCmd (clientData, interp, argc, argv)
    ClientData  clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    request_rec *rr=NULL;
    request_rec *p;
    request_rec *r = Tcl_request_rec;
    char *error_fmt = NULL;
    char c;
    int include;
    int virtual;
    Tcl_Interp *slaveInterp;

    if (argc != 3) {
	Tcl_AppendResult (interp, 
	    "wrong # args: should be \"",
	    argv[0],
	    " slaveInterp file",
	    (char *) NULL);
	return TCL_ERROR;
    }

    if ((slaveInterp = Tcl_GetSlave (interp, argv[1])) == (Tcl_Interp *)NULL) {
	Tcl_AppendResult (interp, 
	    argv[0],
	    ": unknown slave interpreter '",
	    argv[1],
	    "'",
	    (char *) NULL);
	return TCL_ERROR;
    }

    c = argv[0][0];
    if (c == 'i') {
	include = 1;
	c = argv[0][8];
	if (c == 'f') {
	    virtual = 0;
	} else if (c == 'v') {
	    virtual = 1;
	} else {
	    Tcl_AppendResult (interp, 
		argv[0],
		": not invoked as 'SAFE_include_file' or 'SAFE_include_virtual'",
		(char *) NULL);
	    return TCL_ERROR;
	}
    } else if (c == 'l') {
	include = 0;
	c = argv[0][5];
	if (c == 'f') {
	    virtual = 0;
	} else if (c == 'v') {
	    virtual = 1;
	} else {
	    Tcl_AppendResult (interp, 
		argv[0],
		": not invoked as 'SAFE_load_file' or 'SAFE_load_virtual'",
		(char *) NULL);
	    return TCL_ERROR;
	}
    } else {
	Tcl_AppendResult (interp, 
	    argv[0],
	    ": not invoked as 'SAFE_include_' or 'SAFE_load_'",
	    (char *) NULL);
	return TCL_ERROR;
    }

    if (virtual) {
	rr = sub_req_lookup_uri (argv[2], r);
    } else {
	/* be safe; only files in this directory or below allowed */

	/* memory leak here -- tmp never freed */
	char *tmp = (char *)ckalloc(strlen(argv[2]) + 3);
	sprintf(tmp, "/%s/", argv[2]);

	if (*argv[2] == '/' || strstr(tmp, "/../") != NULL) {
	    Tcl_AppendResult (interp, 
		argv[0],
		": unable to include '",
		argv[2],
		"': illegal filename from '",
		r->filename,
		"'",
		(char *) NULL);
	    return TCL_ERROR;
	}
	rr = sub_req_lookup_file (argv[2], r);
    }
	
    if (rr->status != 200) {
	Tcl_AppendResult (interp,
	    argv[0],
	    ": unable to load or include '",
	    argv[2],
	    "' from '",
	    r->filename,
	    "'",
	    (char *) NULL);
	if (rr != NULL) destroy_sub_req (rr);
	return TCL_ERROR;
    }

    if ((allow_options (rr) & OPT_INCNOEXEC) && rr->content_type
	&& (strncmp (rr->content_type, "text/", 5))) {
	    Tcl_AppendResult (interp,
		argv[0],
		": unable to load or include potential exec '",
		argv[2],
		"' from '",
		r->filename,
		"'",
		(char *) NULL);
	    if (rr != NULL) destroy_sub_req (rr);
	    return TCL_ERROR;
    }

    for (p = r; p != NULL; p = p->main) {
	if (strcmp(p->filename, rr->filename) == 0) break;
    }

    if (p != NULL) {
	Tcl_AppendResult (interp,
	    argv[0],
	    ": recursive load or include of '",
	    argv[2],
	    "' from '",
	    r->filename,
	    "'",
	    (char *) NULL);
	if (rr != NULL) destroy_sub_req (rr);
	return TCL_ERROR;
    }
	
    if (include) {
	if (run_sub_req (rr)) {
	    Tcl_AppendResult (interp,
		argv[0],
		": unable to process include '",
		argv[2],
		"' from '",
		r->filename,
		"'",
		(char *) NULL);
	    if (rr != NULL) destroy_sub_req (rr);
	    return TCL_ERROR;
	}
    } else {
	if (load_sub_req (slaveInterp, rr)) {
	    Tcl_AppendResult (interp,
		argv[0],
		": unable to load '",
		argv[2],
		"' from '",
		r->filename,
		"'<p><pre>",
		Tcl_GetVar(slaveInterp, "errorInfo", TCL_GLOBAL_ONLY),
		"</pre><p>",
		(char *) NULL);
	    if (rr != NULL) destroy_sub_req (rr);
	    return TCL_ERROR;
	}
    }

    if (rr != NULL) destroy_sub_req (rr);
    return TCL_OK;
}


/*-----------------------------------------------------------------------------
 * Neo_AbortPageCmd --
 *     Implements the abort_page TCL command:
 *         abort_page
 *
 * Results:
 *      Standard TCL result.
 *-----------------------------------------------------------------------------
 */
int
Neo_AbortPageCmd (clientData, interp, argc, argv)
    ClientData  clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    if (argc != 1) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                          "\"", (char *) NULL);
        return TCL_ERROR;
    }
    
    fflush (Tcl_request_rec->connection->client);
    abort_connection(Tcl_request_rec->connection);
    return TCL_OK;
}


/*-----------------------------------------------------------------------------
 * Neo_UnescapeUrlCmd --
 *     Implements the unescape_url TCL command:
 *         unescape_url pathname
 *
 * Results:
 *      Standard TCL result.
 *-----------------------------------------------------------------------------
 */
int
Neo_UnescapeUrlCmd (clientData, interp, argc, argv)
    ClientData  clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                          " url\"", (char *) NULL);
        return TCL_ERROR;
    }
    
    if (unescape_url(argv[1]) == OK) {
        interp->result = argv[1];
    }
    return TCL_OK;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Neo_RequestInfoCmd --
 *     Implements the Neo TCL request_info command:
 *         request_info args
 *
 * Results:
 *     Standard TCL results.
 *
 *-----------------------------------------------------------------------------
 */
int
Neo_RequestInfoCmd (clientData, interp, argc, argv)
    ClientData  clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    char intbuf[20];
    char *varName;
    int depth;
    request_rec *r = Tcl_request_rec;

    if (argc != 2 && argc != 4) {
	Tcl_AppendResult (interp, 
	    "wrong # args: should be \"",
	    argv[0],
	    " [next|prev|main depth] arrayVarName\"",
	    (char *) NULL);
	return TCL_ERROR;
    }

    if (argc == 2) {
	varName = argv[1];
    } else {
	varName = argv[3];
	strcpy(interp->result,"0");
	if (Tcl_GetInt (interp, argv[2], &depth) == TCL_ERROR)
	    return TCL_ERROR;
	if (strcmp(argv[1], "prev") == 0) {
	   while (depth--) {
	       if (r->prev) {
		   r = r->prev;
	       } else {
		   return TCL_OK;
	       }
	   }
	} else if (strcmp(argv[1], "next") == 0) {
	   while (depth--) {
	       if (r->next) {
		   r = r->next;
	       } else {
		   return TCL_OK;
	       }
	   }
	} else if (strcmp(argv[1], "main") == 0) {
	   while (depth--) {
	       if (r->main) {
		   r = r->main;
	       } else {
		   return TCL_OK;
	       }
	   }
	} else {
	    Tcl_AppendResult (interp,
		"bad arg: should be \"",
		argv[0],
		" [next|prev|main depth] arrayVarName\"",
		(char *) NULL);
	    return TCL_ERROR;
	}
    }

#define REQUEST_CHAR_VAR(X) if (Tcl_SetVar2(interp, varName, #X, r->X == (char *)NULL ? "" : r->X, TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR

#define REQUEST_INT_VAR(X) {sprintf(intbuf, "%d", r->X); if (Tcl_SetVar2(interp, varName, #X, intbuf, TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;}

    REQUEST_CHAR_VAR(the_request);
    REQUEST_INT_VAR(assbackwards);
    REQUEST_INT_VAR(header_only);
    REQUEST_CHAR_VAR(protocol);
    REQUEST_CHAR_VAR(status_line);
    REQUEST_INT_VAR(status);
    REQUEST_CHAR_VAR(method);
    REQUEST_INT_VAR(method_number);
    REQUEST_INT_VAR(bytes_sent);
    REQUEST_CHAR_VAR(content_type);
    REQUEST_CHAR_VAR(content_encoding);
    REQUEST_CHAR_VAR(content_language);
    REQUEST_INT_VAR(no_cache);
    REQUEST_CHAR_VAR(uri);
    REQUEST_CHAR_VAR(filename);
    REQUEST_CHAR_VAR(path_info);
    REQUEST_CHAR_VAR(args);

    if (Tcl_SetVar2 (interp, varName, "main", r->main ? "1" : "0", TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;
    if (Tcl_SetVar2 (interp, varName, "prev", r->prev ? "1" : "0", TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;
    if (Tcl_SetVar2 (interp, varName, "next", r->next ? "1" : "0", TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;

    sprintf(intbuf, "%d", r->finfo.st_uid);
    if (Tcl_SetVar2 (interp, varName, "file_uid", intbuf, TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;

    interp->result = "1";
    return TCL_OK;
}

int
Tcl_ExtendSafeSlaveCmd(dummy, interp, argc, argv)
    ClientData dummy;                   /* Not used. */
    Tcl_Interp *interp;                 /* Current interpreter. */
    int argc;                           /* Number of arguments. */
    char **argv;                        /* Argument strings. */
{   
    Tcl_Interp *slaveInterp;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
	    " slaveInterpName\"", (char *)NULL);
	return TCL_ERROR;
    }
    slaveInterp = Tcl_GetSlave(interp, argv[1]);
    if (slaveInterp == (Tcl_Interp *)NULL) {
	return TCL_ERROR;
    }

    /* Do this in the setup using load so that unsafe version can be used */
    /*
    if (Tclx_SafeInit (slaveInterp) == TCL_ERROR) {
	return TCL_ERROR;
    }
    */

    Tcl_CreateCommand (slaveInterp, "www_request_info", 
		      Neo_RequestInfoCmd,
                      (ClientData) NULL, (void (*)()) NULL);

    return TCL_OK;
}


/*-----------------------------------------------------------------------------
 * Neo_SimplifyPathnameCmd --
 *     Implements the simplify_pathname TCL command:
 *         simplify_pathname pathname
 *
 * Results:
 *      Standard TCL result.
 *-----------------------------------------------------------------------------
 */
int
Neo_SimplifyPathnameCmd (clientData, interp, argc, argv)
    ClientData  clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                          " pathname\"", (char *) NULL);
        return TCL_ERROR;
    }
    
    getparents(argv[1]);
    interp->result = argv[1];
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_HtmlCmd --
 *
 *	This procedure is invoked to process the "html" Tcl command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_HtmlCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    int argIndex;

    if (argc < 2 || argc > 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
	    " text [tag]\"", (char *)NULL);
	return TCL_ERROR;
    }

    if (argc == 2) {
	rprintf (Tcl_request_rec, "%s", argv[1]);
    } else {
	rprintf (Tcl_request_rec, "<%s>%s</%s>", argv[2], argv[1], argv[2]);
    }
    return TCL_OK;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Neo_DigitToHex
 *     Helper function to convert a number 0 - 15 into the equivalent hex
 *     character.
 *
 * Results:
 *     The integer, or -1 if an illegal hex character is encountered.
 *
 *-----------------------------------------------------------------------------
 */
int 
Neo_DigitToHex(int c) {

    if (c < 10) {
        return c + '0';
    }
    return c - 10 + 'a';
}


/*
 *-----------------------------------------------------------------------------
 *
 * Neo_EscapeStringCmd --
 *     Implements the Neo TCL escape_string command:
 *         escape_string string
 *
 * Results:
 *     Standard TCL results.
 *
 *-----------------------------------------------------------------------------
 */
int
Neo_EscapeStringCmd (clientData, interp, argc, argv)
    ClientData  clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    char *origString, *newString, *origStringP, *newStringP, *checkP;
    int origLength;

    if (argc != 2) {
	Tcl_AppendResult (interp, 
	    "wrong # args: should be \"",
	    argv[0],
	    " string",
	    (char *) NULL);
	return TCL_ERROR;
    }

    origString = argv[1];

    /* if they sent us an empty string, we're done */
    origLength = strlen (origString);
    if (origLength == 0) {
	return TCL_OK;
    }
    newString = ckalloc (strlen (origString) * 3 + 1);

    /* for all the characters in the source string */
    for (origStringP = origString, newStringP = newString;
	*origStringP != '\0';
	origStringP++) {
	char c = *origStringP;

        if (isalnum (c)) {
	    *newStringP++ = c;
	} else {
	    *newStringP++ = '%';
	    *newStringP++ = Neo_DigitToHex((c >> 4) & 0x0f);
	    *newStringP++ = Neo_DigitToHex(c & 0x0f);
	}
    }
    /* Don't forget to null-terminate the target string */
    *newStringP = '\0';
    Tcl_SetResult (interp, newString, TCL_DYNAMIC);
    return TCL_OK;
}


/* ----------------------- Initialization function ------------------------- */

void init_neoincludes (server_rec *s, pool *p)
{
    time_t date;
    table *t;
    table_entry *elts;
    int i;
    neoscript_server_config *ns = (neoscript_server_config *)get_module_config(s->module_config ,&includes_and_neoscript_module);


    interp = Tcl_CreateInterp();

    /* get the current time (startup time)
     * so we can return server uptime info */

    time(&date);
    sprintf(softwareStartTimeString, "%ld", date);

    /* Initialize core Tcl components and extensions */

    /* Tcl */
    if (Tcl_Init (interp) == TCL_ERROR) {
	fprintf(stderr, "failed to init NeoWebScript tcl component: %s\n", interp->result);
	exit(1);
    }

    /* Extended Tcl */
    if (Tclx_Init (interp) == TCL_ERROR) {
	fprintf(stderr, "failed to init NeoWebScript tclx component: %s\n", interp->result);
	exit(1);
    }

#ifdef MIT_OTCL
    /* MIT Object Tcl */
    if (Otcl_Init (interp) == TCL_ERROR) {
	fprintf(stderr, "failed to init NeoWebScript otcl component: %s\n", interp->result);
	exit(1);
    }
    Tcl_StaticPackage (interp, "Otcl", Otcl_Init, Otcl_Init);
#endif

#ifdef POSTGRES95
    /* Postgres95 Database Interface */
    if (Pg_Init (interp) == TCL_ERROR) {
	fprintf(stderr, "failed to init NeoWebScript postgres component: %s\n", interp->result);
	exit(1);
    }
    Tcl_StaticPackage (interp, "Pg", Pg_Init, Pg_Init);
#endif

    /* NeoSoft Extensions */
    if (Neo_Init (interp) == TCL_ERROR) {
	fprintf(stderr, "failed to init NeoWebScript neo component: %s\n", interp->result);
	exit(1);
    }
    Tcl_StaticPackage (interp, "Neo", Neo_Init, NULL);

    /*
     * copy any variables defined with neoscript server config commands
     * into a Tcl array
     */

    t = ns->neowebscript_server_vars; 
    elts = (table_entry *)t->elts;

    for (i = 0; i < t->nelts; ++i)
	Tcl_SetVar2 (interp, "NeoWebServerConf", elts[i].key, elts[i].val, TCL_GLOBAL_ONLY);



    /* Tcl_InitMath (interp); */
    Tcl_InitExtensions (interp);

    Tcl_SetVar2 (interp, "server", "SERVER_ROOT", server_root, TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "server", "SERVER_CONF", server_confname, TCL_GLOBAL_ONLY);

    if (Tcl_VarEval(interp, "source ", server_root, "/conf/init.tcl", (char *)NULL) == TCL_ERROR) {
	char *errorInfo;

	errorInfo = Tcl_GetVar (interp, "errorInfo", TCL_GLOBAL_ONLY);
	fprintf(stderr,"NeoWebScript startup failed: %s\n", errorInfo);
	exit(1);
    }
}

void
Tcl_InitExtensions (Tcl_Interp *interp) {

    Tcl_CreateCommand(interp, "SAFE_include_file", Neo_IncludeCmd, 
	(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "SAFE_include_virtual", Neo_IncludeCmd, 
	(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "SAFE_load_file", Neo_IncludeCmd, 
	(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "SAFE_load_virtual", Neo_IncludeCmd, 
	(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "abort_page", Neo_AbortPageCmd, 
	(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "extend_slave", Tcl_ExtendSafeSlaveCmd, 
	(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "html", Tcl_HtmlCmd,
	(ClientData) 0, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand (interp, "www_unescape_string", Neo_UnescapeStringCmd,
                      (ClientData) NULL, (void (*)()) NULL);

    Tcl_CreateCommand (interp, "www_escape_string", 
		      Neo_EscapeStringCmd,
                      (ClientData) NULL, (void (*)()) NULL);

    Tcl_CreateCommand (interp, "www_simplify_pathname", 
		      Neo_SimplifyPathnameCmd,
                      (ClientData) NULL, (void (*)()) NULL);

    Tcl_CreateCommand (interp, "www_unescape_url", 
		      Neo_UnescapeUrlCmd,
                      (ClientData) NULL, (void (*)()) NULL);

    Tcl_CreateCommand (interp, "www_request_info", 
		      Neo_RequestInfoCmd,
                      (ClientData) NULL, (void (*)()) NULL);

    /* TclX_InitDB(interp); */
}

/* ------------------------ Environment function -------------------------- */

void add_include_vars(request_rec *r, char *timefmt)
{
    table *e = r->subprocess_env;
    char *t;
    time_t date = time(NULL);

    table_set(e, "DATE_LOCAL", ht_time(r->pool, date, timefmt, 0));
    table_set(e, "DATE_GMT", ht_time(r->pool, date, timefmt, 1));
    table_set(e, "LAST_MODIFIED",ht_time(r->pool,r->finfo.st_mtime,timefmt,0));
    table_set(e, "DOCUMENT_URI", r->uri);
    table_set(e, "DOCUMENT_PATH_INFO", r->path_info);
    if((t = strrchr(r->filename, '/')))
        table_set (e, "DOCUMENT_NAME", ++t);
    else
        table_set (e, "DOCUMENT_NAME", r->uri);
}

#define GET_CHAR(f,c,r) \
 { \
   int i = getc(f); \
   if(feof(f) || ferror(f) || (i == -1)) { \
        fclose(f); \
        return r; \
   } \
   c = (char)i; \
 }

/* --------------------------- Parser functions --------------------------- */

static char *http2env(pool *a, char *w)
{
    char *res = pstrcat (a, "HTTP_", w, NULL);
    char *cp = res; 
 
    while (*++cp)
      if (*cp == '-') *cp = '_';
      else *cp = toupper(*cp); 
 
    return res;
}
 
/* Grrrr... rputc makes this slow as all-get-out.  Elsewhere, it doesn't
 * matter much, but this is an inner loop...
 */

int find_string(FILE *in,char *str, request_rec *r) {
    int x,l=strlen(str),p;
    char c;

    p=0;
    while(1) {
        GET_CHAR(in,c,1);
        if(c == str[p]) {
            if((++p) == l)
                return 0;
        }
        else {
            if(r) {
                if(p) {
                    for(x=0;x<p;x++) {
                        rputc(str[x],r);
                    }
                }
                rputc(c,r);
            }
            p=0;
        }
    }
}

/*
 * decodes a string containing html entities or numeric character references.
 * 's' is overwritten with the decoded string.
 * If 's' is syntatically incorrect, then the followed fixups will be made:
 *   unknown entities will be left undecoded;
 *   references to unused numeric characters will be deleted.
 *   In particular, &#00; will not be decoded, but will be deleted.
 *
 * drtr
 */

/* maximum length of any ISO-LATIN-1 HTML entity name. */
#define MAXENTLEN (6)

/* The following is a shrinking transformation, therefore safe. */

static void
decodehtml(char *s)
{
    int val, i, j;
    char *p=s;
    char *ents;
    static char *entlist[MAXENTLEN+1]={
	NULL,  /* 0 */
	NULL,  /* 1 */
	"lt\074gt\076", /* 2 */
	"amp\046ETH\320eth\360", /* 3 */
	"quot\042Auml\304Euml\313Iuml\317Ouml\326Uuml\334auml\344euml\353\
iuml\357ouml\366uuml\374yuml\377", /* 4 */
	"Acirc\302Aring\305AElig\306Ecirc\312Icirc\316Ocirc\324Ucirc\333\
THORN\336szlig\337acirc\342aring\345aelig\346ecirc\352icirc\356ocirc\364\
ucirc\373thorn\376", /* 5 */
	"Agrave\300Aacute\301Atilde\303Ccedil\307Egrave\310Eacute\311\
Igrave\314Iacute\315Ntilde\321Ograve\322Oacute\323Otilde\325Oslash\330\
Ugrave\331Uacute\332Yacute\335agrave\340aacute\341atilde\343ccedil\347\
egrave\350eacute\351igrave\354iacute\355ntilde\361ograve\362oacute\363\
otilde\365oslash\370ugrave\371uacute\372yacute\375" /* 6 */
    };

    for (; *s != '\0'; s++, p++) {
	if (*s != '&') {
	    *p = *s;
	    continue;
	}
	/* find end of entity */
	for (i=1; s[i] != ';' && s[i] != '\0'; i++)
	    continue;

	if (s[i] == '\0') {	/* treat as normal data */
	    *p = *s;
	    continue;
	}

	/* is it numeric ? */
	if (s[1] == '#') {
	    for (j=2, val=0; j < i && isdigit(s[j]); j++)
		val = val * 10 + s[j] - '0';
	    s += i;
	    if (j < i || val <= 8 || (val >= 11 && val <= 31) ||
		(val >= 127 && val <= 160) || val >= 256)
		p--;  /* no data to output */
	    else
		*p = val;
	} else{
	    j = i-1;
	    if (i-1 > MAXENTLEN || entlist[i-1] == NULL) { /* wrong length */
		*p = '&';
		continue;  /* skip it */
	    }
	    for (ents=entlist[i-1]; *ents != '\0'; ents += i)
		if (strncmp(s+1, ents, i-1) == 0) break;

	    if (*ents == '\0')
		*p = '&';  /* unknown */
	    else {
		*p = ((const unsigned char *)ents)[i-1];
		s += i;
	    }
	}
    }

    *p = '\0';
}

/*
 * extract the next tag name and value.
 * if there are no more tags, set the tag name to 'done'
 * the tag value is html decoded if dodecode is non-zero
 */

static char *
get_tag(pool *p, FILE *in, char *tag, int tagbuf_len, int dodecode) {
    char *t = tag, *tag_val, c, term;
    int n;

    n = 0;

    do { /* skip whitespace */
	GET_CHAR(in,c,NULL);
    } while (isspace(c));

    /* tags can't start with - */
    if(c == '-') {
        GET_CHAR(in,c,NULL);
        if(c == '-') {
            do {
		GET_CHAR(in,c,NULL);
	    } while (isspace(c));
            if(c == '>') {
                strcpy(tag,"done");
                return tag;
            }
        }
	return NULL; /* failed */
    }

    /* find end of tag name */
    while(1) {
        if(++n == tagbuf_len) {
            t[tagbuf_len - 1] = '\0';
            return NULL;
        }
	if(c == '=' || isspace(c)) break;
	*(t++) = tolower(c);
        GET_CHAR(in,c,NULL);
    }

    *t++ = '\0';
    tag_val = t;

    while (isspace(c)) GET_CHAR(in, c, NULL); /* space before = */
    if (c != '=') return NULL;

    do {
	GET_CHAR(in,c,NULL);  /* space after = */
    } while (isspace(c));

    /* we should allow a 'name' as a value */
    
    if (c != '"' && c != '\'') return NULL;
    term = c;
    while(1) {
	GET_CHAR(in,c,NULL);
	if(++n == tagbuf_len) {
	    t[tagbuf_len - 1] = '\0';
	    return NULL;
	}
	if (c == term) break;
	*(t++) = c;
    }
    *t = '\0';
    if (dodecode) decodehtml(tag_val);
    return pstrdup (p, tag_val);
}

static int
get_directive(FILE *in, char *d) {
    char c;

    /* skip initial whitespace */
    while(1) {
        GET_CHAR(in,c,1);
        if(!isspace(c))
            break;
    }
    /* now get directive */
    while(1) {
        *d++ = tolower(c);
        GET_CHAR(in,c,1);
        if(isspace(c))
            break;
    }
    *d = '\0';
    return 0;
}

/* ------------------------ Environment function -------------------------- */

void propagate_vars_to_neoscript(Tcl_Interp *interp, request_rec *r)
{
    table *e = r->subprocess_env;
    server_rec *s = r->server;
    conn_rec *c = r->connection;
    char *t;
    char timeTextBuf[16];
    
    char port[40];
    
    array_header *hdrs_arr = table_elts (r->headers_in);
    table_entry *hdrs = (table_entry *)hdrs_arr->elts;
    int i;
    
    /* First, add environment vars from headers... this is as per
     * CGI specs, though other sorts of scripting interfaces see
     * the same vars...
     */

    Tcl_UnsetVar (interp, "webenv", TCL_GLOBAL_ONLY);
    
    for (i = 0; i < hdrs_arr->nelts; ++i) {
        if (!hdrs[i].key) continue;

	/* A few headers are special cased --- Authorization to prevent
	 * rogue scripts from capturing passwords; content-type and -length
	 * for no particular reason.
	 */
	
	if (!strcasecmp (hdrs[i].key, "Content-type")) 
	    Tcl_SetVar2 (interp, "webenv", "CONTENT_TYPE", hdrs[i].val, TCL_GLOBAL_ONLY);
	else if (!strcasecmp (hdrs[i].key, "Content-length"))
	    Tcl_SetVar2 (interp, "webenv", "CONTENT_LENGTH", hdrs[i].val, TCL_GLOBAL_ONLY);
	else if (!strcasecmp (hdrs[i].key, "Authorization"))
	    continue;
	else
	    Tcl_SetVar2 (interp, "webenv", http2env (r->pool, hdrs[i].key), hdrs[i].val, TCL_GLOBAL_ONLY);
    }
    
    Tcl_SetVar2 (interp, "webenv", "SERVER_SOFTWARE", SERVER_VERSION, TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "webenv", "SERVER_ADMIN", s->server_admin, TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "webenv", "SERVER_NAME", s->server_hostname, TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "webenv", "NEOSCRIPT_VERSION", NEOSCRIPT_VERSION, TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "webenv", "NEO_SOFTWARE_START", softwareStartTimeString, TCL_GLOBAL_ONLY);
    sprintf(port, "%d", s->port);
    Tcl_SetVar2 (interp, "webenv", "SERVER_PORT", port, TCL_GLOBAL_ONLY);

    Tcl_SetVar2 (interp, "webenv", "SERVER_ROOT", server_root, TCL_GLOBAL_ONLY);


    Tcl_SetVar2 (interp, "webenv", "REMOTE_HOST", c->remote_name, TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "webenv", "REMOTE_ADDR", c->remote_ip, TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "webenv", "DOCUMENT_ROOT", document_root(r), TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "webenv", "SCRIPT_FILENAME", r->filename, TCL_GLOBAL_ONLY);
    

    if (c->user) Tcl_SetVar2(interp, "webenv", "REMOTE_USER", c->user, TCL_GLOBAL_ONLY);
    if (c->auth_type) Tcl_SetVar2(interp, "webenv", "AUTH_TYPE", c->auth_type, TCL_GLOBAL_ONLY);
    if (c->remote_logname) Tcl_SetVar2(interp, "webenv", "REMOTE_IDENT", c->remote_logname, TCL_GLOBAL_ONLY);
    
    /* Apache custom error responses. If we have redirected set two new vars */
    if (r->prev) {
        if (r->prev->args) Tcl_SetVar2(interp, "webenv","REDIRECT_QUERY_STRING", r->prev->args, TCL_GLOBAL_ONLY);
	if (r->prev->uri) Tcl_SetVar2 (interp, "webenv", "REDIRECT_URL", r->prev->uri, TCL_GLOBAL_ONLY);
    }

    /* these four are normally for CGI's */
    Tcl_SetVar2 (interp, "webenv", "GATEWAY_INTERFACE", "CGI/1.1", TCL_GLOBAL_ONLY);  
    Tcl_SetVar2 (interp, "webenv", "SERVER_PROTOCOL", r->protocol, TCL_GLOBAL_ONLY);  
    Tcl_SetVar2 (interp, "webenv", "REQUEST_METHOD", r->method, TCL_GLOBAL_ONLY);  
    Tcl_SetVar2 (interp, "webenv", "DOCUMENT_URI", r->uri, TCL_GLOBAL_ONLY);
    if((t = strrchr(r->filename, '/')))
        Tcl_SetVar2 (interp, "webenv", "DOCUMENT_NAME", ++t, TCL_GLOBAL_ONLY);
    else
        Tcl_SetVar2 (interp, "webenv", "DOCUMENT_NAME", r->uri, TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "webenv", "DOCUMENT_PATH_INFO", r->path_info, TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, "webenv", "QUERY_STRING", r->args ? r->args : "", TCL_GLOBAL_ONLY);  
    sprintf (timeTextBuf, "%ld", r->finfo.st_mtime);
    Tcl_SetVar2 (interp, "webenv", "NEO_LAST_MODIFIED", timeTextBuf, TCL_GLOBAL_ONLY);

    sprintf (timeTextBuf, "%ld", r->finfo.st_uid);
    Tcl_SetVar2 (interp, "webenv", "NEO_DOCUMENT_UID", timeTextBuf, TCL_GLOBAL_ONLY);

    Tcl_SetVar2 (interp, "webenv", "NEO_TIME_FORMAT", DEFAULT_TIME_FORMAT, TCL_GLOBAL_ONLY);
}

/* --------------------------- Action handlers ---------------------------- */

int run_pickfile_req (request_rec *r)
{
    int errstatus; 
    FILE *f;
    char buf[IOBUFSIZE];
    int nLines;
    int whichLine;
 
    if (r->method_number != M_GET) return DECLINED;
    if (r->finfo.st_mode == 0 || (r->path_info && *r->path_info)) {
        log_reason("File does not exist", r->filename, r);
        return NOT_FOUND;
    } 
  
    if ((errstatus = set_content_length (r, r->finfo.st_size)) 
        || (errstatus = set_last_modified (r, r->finfo.st_mtime)))  
        return errstatus;
  
    f = fopen (r->filename, "r"); 
 
    if (f == NULL) {
        log_reason("file permissions deny server access", r->filename, r);
        return FORBIDDEN;
    }

    if ((fgets(buf,IOBUFSIZE,f)) == NULL) {
	if (errno != EINTR) goto bye;
    }
    nLines = atoi(buf);
    if (nLines <= 0) return OK;
 
    soft_timeout ("send-pick", r);

    srand((int)(getpid() * 17 + time((long *) 0)));
    for (whichLine = rand() % nLines; whichLine-- >= 0; ) {
	if ((fgets(buf,IOBUFSIZE,f)) == NULL) {
	    if (errno != EINTR) break;
	}
    }
    rprintf (r, "%s", buf);
    bye:
    fclose (f);
    return OK;
}
 
int include_cgi(char *s, request_rec *r)
{
    request_rec *rr = sub_req_lookup_uri (s, r);
    
    if (rr->status != 200) return -1;
    
    /* No hardwired path info or query allowed */
    
    if ((rr->path_info && rr->path_info[0]) || rr->args) return -1;
    if (rr->finfo.st_mode == 0) return -1;

    /* Script gets parameters of the *document*, for back compatibility */
    
    rr->path_info = r->path_info; /* painful to get right; see mod_cgi.c */
    rr->args = r->args;
    
    /* Force sub_req to be treated as a CGI request, even if ordinary
     * typing rules would have called it something else.
     */

    rr->content_type = CGI_MAGIC_TYPE;

    /* Run it. */
    
    if (run_sub_req (rr) == REDIRECT) {
        char *location = table_get (rr->headers_out, "Location");
	location = escape_html(rr->pool, location);
	rprintf(r,"<A HREF=\"%s\">%s</A>",location,location);
    }
    
    destroy_sub_req (rr);
    
    return 0;
}

int handle_include(FILE *in, request_rec *r, char *error, int noexec) {
    char tag[MAX_STRING_LEN];
    char *tag_val;

    while(1) {
        if(!(tag_val = get_tag(r->pool, in, tag, MAX_STRING_LEN, 1)))
            return 1;
        if(!strcmp(tag,"file") || !strcmp (tag, "virtual")) {
	    request_rec *rr=NULL;
	    char *error_fmt = NULL;

	    if (tag[0] == 'f')
	    { /* be safe; only files in this directory or below allowed */
		char tmp[MAX_STRING_LEN+2];
		sprintf(tmp, "/%s/", tag_val);
		if (tag_val[0] == '/' || strstr(tmp, "/../") != NULL)
		    error_fmt = "unable to include file %s in parsed file %s";
		else
		    rr = sub_req_lookup_file (tag_val, r);
	    } else
		rr = sub_req_lookup_uri (tag_val, r);
	    
	    if (!error_fmt && rr->status != 200)
	        error_fmt = "unable to include %s in parsed file %s";

	    if (!error_fmt && noexec && rr->content_type
		&& (strncmp (rr->content_type, "text/", 5)))
	        error_fmt =
		  "unable to include potential exec %s in parsed file %s";

	    if (error_fmt == NULL)
	    {
		request_rec *p;

		for (p=r; p != NULL; p=p->main)
		    if (strcmp(p->filename, rr->filename) == 0) break;
		if (p != NULL)
		    error_fmt = "Recursive include of %s in parsed file %s";
	    }
	    
	    if (!error_fmt && run_sub_req (rr))
	        error_fmt = "unable to include %s in parsed file %s";
		    
            if (error_fmt) {
		log_printf(r->server, error_fmt, tag_val, r->filename);
                rprintf(r,"%s",error);
            }            

	    if (rr != NULL) destroy_sub_req (rr);
        } 
        else if(!strcmp(tag,"pickfile") || !strcmp (tag, "pickvirtual")) {
	    request_rec *rr=NULL;
	    char *error_fmt = NULL;

	    if (tag[4] == 'f')
	    { /* be safe; only files in this directory or below allowed */
		char tmp[MAX_STRING_LEN+2];
		sprintf(tmp, "/%s/", tag_val);
		if (tag_val[0] == '/' || strstr(tmp, "/../") != NULL)
		    error_fmt = "unable to include file %s in parsed file %s";
		else
		    rr = sub_req_lookup_file (tag_val, r);
	    } else
		rr = sub_req_lookup_uri (tag_val, r);
	    
	    if (!error_fmt && rr->status != 200)
	        error_fmt = "unable to include %s in parsed file %s";

	    if (!error_fmt && noexec && rr->content_type
		&& (strncmp (rr->content_type, "text/", 5)))
	        error_fmt =
		  "unable to include potential exec %s in parsed file %s";
	    
	    if (!error_fmt && run_pickfile_req (rr))
	        error_fmt = "unable to pick from %s in parsed file %s";
		    
            if (error_fmt) {
                log_printf(r->server, error_fmt, tag_val, r->filename);
                rprintf(r,"%s",error);
            }            

	    if (rr != NULL) destroy_sub_req (rr);
        } 
        else if(!strcmp(tag,"counter")) {
	    extern int errno;
	    DBT key;
	    DBT data;
	    char newNumberText[16];
	    int count;
	    char *error_fmt = NULL;
	    /* char *dbfname = server_root_relative(r->pool, r->server->url_counter_db); */
	    char *dbfname = server_root_relative(r->pool, URL_ACCESS_COUNTER_DB);
	    DB *db = dbopen(dbfname, (O_CREAT|O_RDWR|DB_LOCK), 0644, DB_HASH, (void *)NULL);
	    
	    if (db == (DB *)NULL) {
		error_fmt = "[%s: %s]";
		log_printf (r->server, error_fmt, dbfname, strerror(errno));
		/* rprintf(r, "%s", error); */
		rprintf(r, error_fmt, dbfname, strerror(errno));
		goto countfail;
	    }
	 
	    key.data = r->filename;
	    key.size = strlen(r->filename) + 1;
	   
	    if (db->get(db, &key, &data, 0) != 0) { 
		count = 0; 
	    } else {
		count = atoi(data.data);
	    }
	    sprintf(newNumberText, "%d", ++count);
	    data.data = newNumberText;
	    data.size = strlen(newNumberText) + 1;

	    if (db->put(db, &key, &data, 0) != 0) {
		db->close(db);
		log_printf(r->server, "dbput failed! %s in %s, errno=%d",key,data,errno);
		rprintf(r,"%s",error);
	    }
	    if (db->sync(db, 0) < 0) {
		log_printf (r->server, "dbsync failed! errno=%d", errno);
		rprintf(r,"%s",error);
	    }
	    if (db->close(db) < 0) {
		log_printf (r->server, "dbclose failed! errno=%d", errno);
		rprintf(r,"%s",error);
	    }
	    rprintf (r, "%s", newNumberText);
	    countfail: ;
	}
        else if(!strcmp(tag,"done"))
            return 0;
        else {
            log_printf(r->server, "unknown parameter '%s' to tag 'include' in %s",
		       tag, r->filename);
            rprintf(r,"%s",error);
        }
    }
}

typedef struct {
    request_rec *r;
    char *s;
} include_cmd_arg;

void include_cmd_child (void *arg)
{
    request_rec *r =  ((include_cmd_arg *)arg)->r;
    char *s = ((include_cmd_arg *)arg)->s;
    table *env = r->subprocess_env;
#ifdef DEBUG_NEOINCLUDE_CMD    
    FILE *dbg = fopen ("/dev/tty", "w");
#endif    
    char err_string [MAX_STRING_LEN];

#ifdef DEBUG_NEOINCLUDE_CMD    
    fprintf (dbg, "Attempting to neoinclude command '%s'\n", s);
#endif    

    if (r->path_info && r->path_info[0] != '\0')
    {
	request_rec *pa_req;

	table_set (env, "PATH_INFO", escape_shell_cmd (r->pool, r->path_info));
	
	pa_req = sub_req_lookup_uri(escape_uri(r->pool, r->path_info), r);
	if (pa_req->filename)
	    table_set(env, "PATH_TRANSLATED",
		      pstrcat(r->pool, pa_req->filename, pa_req->path_info,
			      NULL));
    }

    if (r->args) {
        table_set (env, "QUERY_STRING", r->args);
	unescape_url (r->args);
	table_set (env, "QUERY_STRING_UNESCAPED",
		   escape_shell_cmd (r->pool, r->args));
    }
    
    error_log2stderr (r->server);
    
#ifdef DEBUG_NEOINCLUDE_CMD    
    fprintf (dbg, "Attempting to exec '%s'\n", s);
#endif    
    cleanup_for_exec();
    execle(SHELL_PATH, SHELL_PATH, "-c", s, NULL,
	   create_environment (r->pool, env));
    
    /* Oh, drat.  We're still here.  The log file descriptors are closed,
     * so we have to whimper a complaint onto stderr...
     */
    
#ifdef DEBUG_NEOINCLUDE_CMD    
    fprintf (dbg, "Exec failed\n");
#endif    
    sprintf(err_string, "httpd: exec of %s failed, errno is %d\n",
	    SHELL_PATH,errno);
    write (2, err_string, strlen(err_string));
    exit(0);
}

int include_cmd(char *s, request_rec *r) {
    include_cmd_arg arg;
    FILE *f;

    arg.r = r; arg.s = s;

    if (!spawn_child (r->connection->pool, include_cmd_child, &arg,
		      kill_after_timeout, NULL, &f))
        return -1;
    
    send_fd(f,r);
    pfclose(r->pool, f);	/* will wait for zombie when
				 * r->pool is cleared
				 */
    return 0;
}

int handle_exec(FILE *in, request_rec *r, char *error)
{
    char tag[MAX_STRING_LEN];
    char *tag_val;
    char *file = r->filename;

    while(1) {
        if(!(tag_val = get_tag (r->pool, in, tag, MAX_STRING_LEN, 1)))
            return 1;
        if(!strcmp(tag,"cmd")) {
            if(include_cmd(tag_val, r) == -1) {
                log_printf(r->server, "failed command exec %s in %s",
			   tag_val, file);
                rprintf(r,"%s",error);
            }
            /* just in case some stooge changed directories */
            chdir_file(r->filename);
        } 
        else if(!strcmp(tag,"cgi")) {
            if(include_cgi(tag_val, r) == -1) {
                log_printf(r->server, "invalid CGI ref %s in %s",tag_val,file);
                rprintf(r,"%s",error);
            }
            /* grumble groan */
            chdir_file(r->filename);
        }
        else if(!strcmp(tag,"done"))
            return 0;
        else {
            log_printf(r->server, "unknown parameter '%s' to tag 'exec' in %s",
		       tag, file);
            rprintf(r, "%s",error);
        }
    }
}

int handle_echo (FILE *in, request_rec *r, char *error) {
    char tag[MAX_STRING_LEN];
    char *tag_val;

    while(1) {
        if(!(tag_val = get_tag (r->pool, in, tag, MAX_STRING_LEN, 1)))
            return 1;
        if(!strcmp(tag,"var")) {
	    char *val = table_get (r->subprocess_env, tag_val);

	    if (val) rprintf (r, "%s", val);
	    else rprintf (r, "(none)");
        } else if(!strcmp(tag,"done"))
            return 0;
        else {
            log_printf(r->server, "unknown parameter '%s' to tag 'echo' in %s",
		    tag, r->filename);
            rprintf (r, "%s", error);
        }
    }
}

int handle_config(FILE *in, request_rec *r, char *error, char *tf,
                  int *sizefmt) {
    char tag[MAX_STRING_LEN];
    char *tag_val;
    table *env = r->subprocess_env;

    while(1) {
        if(!(tag_val = get_tag(r->pool, in, tag, MAX_STRING_LEN, 0)))
            return 1;
        if(!strcmp(tag,"errmsg"))
            strcpy(error,tag_val);
        else if(!strcmp(tag,"timefmt")) {
	    time_t date = time(NULL);
            strcpy(tf,tag_val);
            table_set (env, "DATE_LOCAL", ht_time(r->pool,date,tf,0));
            table_set (env, "DATE_GMT", ht_time(r->pool,date,tf,1));
            table_set (env, "LAST_MODIFIED", ht_time(r->pool,r->finfo.st_mtime,tf,0));
        }
        else if(!strcmp(tag,"sizefmt")) {
	    decodehtml(tag_val);
            if(!strcmp(tag_val,"bytes"))
                *sizefmt = SIZEFMT_BYTES;
            else if(!strcmp(tag_val,"abbrev"))
                *sizefmt = SIZEFMT_KMG;
        } 
        else if(!strcmp(tag,"done"))
            return 0;
        else {
            log_printf(r->server, "unknown parameter '%s' to tag 'config' in %s",
                    tag, r->filename);
            rprintf (r,"%s",error);
        }
    }
}



int find_file(request_rec *r, char *directive, char *tag, 
              char *tag_val, struct stat *finfo, char *error)
{
    char dir[MAX_STRING_LEN];
    char *to_send;

    if(!strcmp(tag,"file")) {
        getparents(tag_val); /* get rid of any nasties */
        getwd(dir);
        to_send = make_full_path (r->pool, dir, tag_val);
        if(stat(to_send,finfo) == -1) {
            log_printf(r->server,
                    "unable to get information about %s in parsed file %s",
                    to_send, r->filename);
            rprintf (r,"%s",error);
            return -1;
        }
        return 0;
    }
    else if(!strcmp(tag,"virtual")) {
	request_rec *rr = sub_req_lookup_uri (tag_val, r);
	
	if (rr->status == 200 && rr->finfo.st_mode != 0) {
	    memcpy ((char*)finfo, (const char *)&rr->finfo, sizeof (struct stat));
	    destroy_sub_req (rr);
	    return 0;
        } else {
            log_printf(r->server,
                    "unable to get information about %s in parsed file %s",
                    tag_val, r->filename);
            rprintf(r,"%s",error);
	    destroy_sub_req (rr);
            return -1;
        }
    }
    else {
        log_printf(r->server, "unknown parameter '%s' to tag '%s' in %s",
                tag, directive, r->filename);
        rprintf(r,"%s",error);
        return -1;
    }
}


int handle_fsize(FILE *in, request_rec *r, char *error, int sizefmt) 
{
    char tag[MAX_STRING_LEN];
    char *tag_val;
    struct stat finfo;

    while(1) {
        if(!(tag_val = get_tag(r->pool, in, tag, MAX_STRING_LEN, 1)))
            return 1;
        else if(!strcmp(tag,"done"))
            return 0;
        else if(!find_file(r,"fsize",tag,tag_val,&finfo,error)) {
            if(sizefmt == SIZEFMT_KMG) {
                send_size(finfo.st_size, r);
            }
            else {
                int l,x;
#if defined(BSD) && BSD > 199305
                sprintf(tag,"%qd",finfo.st_size);
#else
                sprintf(tag,"%ld",finfo.st_size);
#endif
                l = strlen(tag); /* grrr */
                for(x=0;x<l;x++) {
                    if(x && (!((l-x) % 3))) {
                        rputc(',', r);
                    }
                    rputc (tag[x],r);
                }
            }
        }
    }
}

int handle_flastmod(FILE *in, request_rec *r, char *error, char *tf) 
{
    char tag[MAX_STRING_LEN];
    char *tag_val;
    struct stat finfo;

    while(1) {
        if(!(tag_val = get_tag(r->pool, in, tag, MAX_STRING_LEN, 1)))
            return 1;
        else if(!strcmp(tag,"done"))
            return 0;
        else if(!find_file(r,"flastmod",tag,tag_val,&finfo,error))
            rprintf (r, "%s", ht_time(r->pool, finfo.st_mtime, tf, 0));
    }
}    

int handle_neoscript (FILE *in, request_rec *r, char *error) {
    char tag[MAX_STRING_LEN];
    char *tag_val;
    extern Tcl_Interp *interp;
    Tcl_Interp *safeInterp;
    struct request_rec *q;
    char *safeInterpName;
    request_rec *Tcl_saved_request_rec;
    table *t;
    table_entry *elts;
    int i;

    Tcl_saved_request_rec = Tcl_request_rec;
    Tcl_request_rec = r;

    /* look into the request record and find out if we already have a safe 
     * interpreter running with the right user ID.
     *
     * if not, create one and register it.  if so, hook up to it.
     *
     */

     for (q = r, safeInterpName = (char *)NULL; q != (struct request_rec *)NULL; q = q->main) {
	 if (r->finfo.st_uid != q->finfo.st_uid) break;
	 if ((safeInterpName = get_module_config(q->request_config, &includes_and_neoscript_module)) != NULL) break;
     }
     if (safeInterpName) {
	    safeInterp = Tcl_GetSlave(interp, safeInterpName);
	    assert (safeInterp != NULL);
     } else {

	/* Set array for the config information here, once, when the slave
	 * interp is created.  Saves repeating work when several script blocks
	 * exist within a page.
	 */
	neoscript_dir_config *ns = (neoscript_dir_config *)get_module_config(r->per_dir_config,&includes_and_neoscript_module);

	t = ns->neowebscript_dir_vars; 
	elts = (table_entry *)t->elts;

	/* Make sure it doesn't exist from a previous page served by this process */
	Tcl_UnsetVar (interp, "NeoWebDirConf", TCL_GLOBAL_ONLY);
	for (i = 0; i < t->nelts; ++i)
	    Tcl_SetVar2 (interp, "NeoWebDirConf", elts[i].key, elts[i].val, TCL_GLOBAL_ONLY);

	if (Tcl_GlobalEval(interp, "setup_safe_interpreter") != TCL_OK) {
	    char *errorInfo;

	    errorInfo = Tcl_GetVar (interp, "errorInfo", TCL_GLOBAL_ONLY);
            log_error(errorInfo, r->server);
	    fprintf(stderr, "setup_safe_interpreter: %s\n", errorInfo);
	    exit(1);
	} else {
	    safeInterpName = (char *)palloc (r->pool, strlen(interp->result) + 1);
	    strcpy (safeInterpName, interp->result);

	    safeInterp = Tcl_GetSlave(interp, safeInterpName);
	    /*assert (safeInterp != NULL);*/
	    if (safeInterp == NULL) {
		exit(1);
	    }
	}

	 register_cleanup (r->pool, (void *)safeInterp, (void (*)())Tcl_DeleteInterp, (void (*)())Tcl_DeleteInterp);
	 set_module_config(r->request_config, &includes_and_neoscript_module, safeInterpName);

	propagate_vars_to_neoscript (safeInterp, r);
    }

    /* parse the tag=value pairs.  build up a neoscript command
     * including name of the safe interp we're running against.
     */
    while(1) {
	Tcl_DString userCommand;
	char *commandString;

        if(!(tag_val = get_tag (r->pool, in, tag, MAX_STRING_LEN, 1))) {
    	    Tcl_request_rec = Tcl_saved_request_rec;
            return 1;
	}

        if(!strcmp(tag,"done"))
	    goto done_ok;

	Tcl_DStringInit(&userCommand);
	Tcl_DStringAppendElement(&userCommand, "handle_neoscript_request");
	Tcl_DStringAppendElement(&userCommand, safeInterpName);
	Tcl_DStringAppendElement(&userCommand, tag);
	Tcl_DStringAppendElement(&userCommand, tag_val);
	commandString = Tcl_DStringValue(&userCommand);

	if (Tcl_GlobalEval (interp, commandString) == TCL_ERROR) {
	    rprintf (r, "[%s error %s]", tag, interp->result);
	}
    }
done_ok:
    Tcl_request_rec = Tcl_saved_request_rec;
    return 0;
}

int handle_neoscript_config(FILE *in, request_rec *r, char *error, char *tf,
                  int *sizefmt) {
    char tag[MAX_STRING_LEN];
    char *tag_val;
    table *env = r->subprocess_env;

    while(1) {
        if(!(tag_val = get_tag(r->pool, in, tag, MAX_STRING_LEN, 0)))
            return 1;
        if(!strcmp(tag,"errmsg"))
            strcpy(error,tag_val);
        else if(!strcmp(tag,"timefmt")) {
	    time_t date = time(NULL);
            strcpy(tf,tag_val);
            table_set (env, "DATE_LOCAL", ht_time(r->pool,date,tf,0));
            table_set (env, "DATE_GMT", ht_time(r->pool,date,tf,1));
            table_set (env, "LAST_MODIFIED", ht_time(r->pool,r->finfo.st_mtime,tf,0));
        }
        else if(!strcmp(tag,"sizefmt")) {
	    decodehtml(tag_val);
            if(!strcmp(tag_val,"bytes"))
                *sizefmt = SIZEFMT_BYTES;
            else if(!strcmp(tag_val,"abbrev"))
                *sizefmt = SIZEFMT_KMG;
        } 
        else if(!strcmp(tag,"done"))
            return 0;
        else {
            char errstr[MAX_STRING_LEN];
            sprintf(errstr,"unknown parameter '%s' to tag 'config' in %s",
                    tag, r->filename);
            log_error(errstr, r->server);
            rprintf (r,"%s",error);
        }
    }
}




/* -------------------------- The main function --------------------------- */

/* This is a stub which parses a file descriptor. */

void send_parsed_content(FILE *f, request_rec *r)
{
    char directive[MAX_STRING_LEN], error[MAX_STRING_LEN];
    char timefmt[MAX_STRING_LEN];
    int noexec = allow_options (r) & OPT_INCNOEXEC;
    int ret, sizefmt;

    strcpy(error,DEFAULT_ERROR_MSG);
    strcpy(timefmt,DEFAULT_TIME_FORMAT);
    sizefmt = SIZEFMT_KMG;

    chdir_file (r->filename);

    while(1) {
        if(!find_string(f,STARTING_SEQUENCE,r)) {
            if(get_directive(f,directive))
                return;
            if(!strcmp(directive,"exec")) {
                if(noexec) {
                    log_printf(r->server,
			       "httpd: exec used but not allowed in %s",
			       r->filename);
                    rprintf(r,"%s",error);
                    ret = find_string(f,ENDING_SEQUENCE,NULL);
                } else 
                    ret=handle_exec(f, r, error);
            } 
            else if (!strcmp(directive,"neowebscript") || !strcmp(directive,"neoscript"))
                ret=handle_neoscript(f, r, error);
            else if(!strcmp(directive,"neoconfig"))
                ret=handle_neoscript_config(f, r, error, timefmt, &sizefmt);
            else if(!strcmp(directive,"config"))
                ret=handle_config(f, r, error, timefmt, &sizefmt);
            else if(!strcmp(directive,"include"))
                ret=handle_include(f, r, error, noexec);
            else if(!strcmp(directive,"echo"))
                ret=handle_echo(f, r, error);
            else if(!strcmp(directive,"fsize"))
                ret=handle_fsize(f, r, error, sizefmt);
            else if(!strcmp(directive,"flastmod"))
                ret=handle_flastmod(f, r, error, timefmt);
            else {
                log_printf(r->server, 
			   "httpd: unknown directive '%s' in parsed doc %s",
			   directive, r->filename);
                rprintf (r,"%s",error);
                ret=find_string(f,ENDING_SEQUENCE,NULL);
            }
            if(ret) {
                log_printf(r->server, "httpd: premature EOF in parsed file %s",
			   r->filename);
                return;
            }
        } else 
            return;
    }
}

char *set_xbithack (cmd_parms *cmd, neoscript_dir_config *m, char *arg)
{
   if (!strcasecmp (arg, "off")) m->xbithack = xbithack_off;
   else if (!strcasecmp (arg, "on")) m->xbithack = xbithack_on;
   else if (!strcasecmp (arg, "full")) m->xbithack = xbithack_full;
   else return "XBitHack must be set to Off, On, or Full";

   return NULL;
}

int send_parsed_file(request_rec *r)
{
    FILE *f;
    neoscript_dir_config *ns = (neoscript_dir_config *)get_module_config(r->per_dir_config,&includes_and_neoscript_module);
    enum xbithack xbit = ns->xbithack;
    int errstatus;

    extern Tcl_Interp *interp;
    int is_included = !strcmp (r->protocol, "INCLUDED");
    char *argv0;
    int nph;
    char *lenp = table_get (r->headers_in, "Content-length");
    char argsbuffer[HUGE_STRING_LEN];

    if (!(allow_options (r) & OPT_INCLUDES)) return DECLINED;

    /* if (r->method_number != M_GET) return DECLINED; */

    if (r->finfo.st_mode == 0) return NOT_FOUND;
	
    if (xbit == xbithack_full
	&& (r->finfo.st_mode & S_IXGRP)
	&& (errstatus = set_last_modified (r, r->finfo.st_mtime)))
        return errstatus;
    
    if(!(f=pfopen(r->pool, r->filename, "r"))) {
        log_reason("file permissions deny server access", r->filename, r);
	return FORBIDDEN;
    }
    
    /**** Handle put and post for NeoScript ****/
    if((argv0 = strrchr(r->filename,'/')) != NULL)
        argv0++;
    else argv0 = r->filename;

    nph = !(strncmp(argv0,"nph-",4));
    
    if (nph && is_included) {
        log_reason("attempt to include NPH NeoScript CGI script", r->filename, r);
	return FORBIDDEN;
    }
    
    if ((r->method_number == M_POST || r->method_number == M_PUT)
	&& !lenp) {
        log_reason("POST or PUT without Content-length:", r->filename, r);
	return BAD_REQUEST;
    }
    /**** End of put and post handling for NeoScript ****/

    r->content_type = "text/html";
    
    hard_timeout ("send", r);

    /***** More put and post handling for NeoScript ****/
    if (!nph) {
        send_http_header(r);

        if (r->header_only) {
            kill_timeout (r);
	    pfclose (r->pool, f);
	    return OK;
        }
    }
    /***** End of more put and post handling for NeoScript ****/
   
    if (r->main) {
	/* Kludge --- for nested includes, we want to keep the
	 * subprocess environment of the base document (for compatibility);
	 * that means torquing our own last_modified date as well so that
	 * the LAST_MODIFIED variable gets reset to the proper value if
	 * the nested document resets <!--#config timefmt-->
	 */
	r->subprocess_env = r->main->subprocess_env;
	r->finfo.st_mtime= r->main->finfo.st_mtime;
    } else { 
	add_common_vars (r);
	add_include_vars (r, DEFAULT_TIME_FORMAT);

        /* ??? Not needed most of the time; need to figure this out */
	propagate_vars_to_neoscript (interp, r);
    }
    
    /* Last part of put and post handling */
    /* It's a post or a put, read the data from stdin and build
     * it up in a tcl dstring, then store it into an element
     * of our webenv array.
     *
     * This may not be the best thing to do, but it's a place
     * to begin.
     */
    if (r->method_number == M_POST || r->method_number == M_PUT) {
	Tcl_DString tclStdinString;
        void (*handler)();

	int remaining = atoi (lenp);
	
        hard_timeout ("copy neowebscript args", r);
        handler = signal (SIGPIPE, SIG_IGN);

	Tcl_DStringInit(&tclStdinString);
    
	while ((remaining > 0))
	{
	    int len_read, len_to_read = remaining;

	    if (len_to_read > HUGE_STRING_LEN) len_to_read = HUGE_STRING_LEN;
	    
	    len_read = read_client_block (r, argsbuffer, len_to_read);

	    Tcl_DStringAppend(&tclStdinString, argsbuffer, len_read);
	    remaining -= len_read;
	}
    
	signal (SIGPIPE, handler);
	
	kill_timeout (r);

	Tcl_SetVar2 (interp, "webenv", "NEO_POST_DATA", Tcl_DStringValue(&tclStdinString), TCL_GLOBAL_ONLY);
	Tcl_DStringFree(&tclStdinString);
    }
    /* End of last part of put and post handling for NeoScript */
    
    send_parsed_content (f, r);
    
    kill_timeout (r);
    return OK;
}

int xbithack_handler (request_rec *r)
{
    neoscript_dir_config *ns;
    enum xbithack xbit;

    if (!(r->finfo.st_mode & S_IXUSR)) return DECLINED;

    ns = (neoscript_dir_config *)
	get_module_config(r->per_dir_config, &includes_and_neoscript_module);

    xbit = ns->xbithack;
	
    if (xbit == xbithack_off) return DECLINED;
    return send_parsed_file (r);
}

char *neowebscript_dir_command(cmd_parms *cmd, neoscript_dir_config *m, char *var, char *val)
{
    table_set (m->neowebscript_dir_vars, var, val);
    return NULL;
}

char *neowebscript_server_command(cmd_parms *cmd, void *dummy, char *var, char *val)
{
    server_rec *s = cmd->server;
    neoscript_server_config *ns;

    ns = (neoscript_server_config *)get_module_config(s->module_config, &includes_and_neoscript_module);
    table_set (ns->neowebscript_server_vars, var, val);
    return NULL;
}


/* -------------------------- The main function --------------------------- */

/* This is a stub which parses a file descriptor. */

void *create_neoscript_dir_config (pool *p, char *dummy)
{
    neoscript_dir_config *new =
      (neoscript_dir_config *) palloc (p, sizeof(neoscript_dir_config));

    new->neowebscript_dir_vars = make_table (p, 4);
    new->xbithack = DEFAULT_XBITHACK;

    return new;
}

void *merge_neoscript_dir_configs (pool *p, void *basev, void *addv)
{
    neoscript_dir_config *base = (neoscript_dir_config *)basev;
    neoscript_dir_config *add = (neoscript_dir_config *)addv;
    neoscript_dir_config *new =
      (neoscript_dir_config *)palloc (p, sizeof(neoscript_dir_config));

    new->neowebscript_dir_vars = overlay_tables (p, add->neowebscript_dir_vars,
					base->neowebscript_dir_vars);
    new->xbithack = add->xbithack;

    return new;
}

void *create_neoscript_server_config (pool *p, server_rec *s)
{
    neoscript_server_config *new =
      (neoscript_server_config *) palloc (p, sizeof(neoscript_server_config));

    new->neowebscript_server_vars = make_table (p, 4);

    return new;
}

void *merge_neoscript_server_configs (pool *p, void *basev, void *addv)
{
    neoscript_server_config *base = (neoscript_server_config *)basev;
    neoscript_server_config *add = (neoscript_server_config *)addv;
    neoscript_server_config *new =
      (neoscript_server_config *)palloc (p, sizeof(neoscript_server_config));

    new->neowebscript_server_vars 
	= overlay_tables (p, add->neowebscript_server_vars,
			    base->neowebscript_server_vars);

    return new;
}

command_rec neoincludes_cmds[] = {
{ "XBitHack", set_xbithack, NULL, OR_OPTIONS, TAKE1, "Off, On, or Full" },
{ "NeoWebServerConf", neowebscript_server_command, NULL, RSRC_CONF, TAKE2, NULL },
{ "NeoWebDirConf", neowebscript_dir_command, NULL, ACCESS_CONF, TAKE2, NULL },
{ NULL }    
};

handler_rec neoincludes_handlers[] = {
{ INCLUDES_MAGIC_TYPE, send_parsed_file },
{ INCLUDES_MAGIC_TYPE3, send_parsed_file },
{ "*/*", xbithack_handler },
{ NULL }
};

module includes_and_neoscript_module = {
   STANDARD_MODULE_STUFF,
   init_neoincludes,		/* initializer */
   create_neoscript_dir_config,	/* dir config creater */
   merge_neoscript_dir_configs,	/* dir merger --- default is to override */
   create_neoscript_server_config,	/* server config */
   merge_neoscript_server_configs,	/* merge server config */
   neoincludes_cmds,		/* command table */
   neoincludes_handlers,	/* handlers */
   NULL,			/* filename translation */
   NULL,			/* check_user_id */
   NULL,			/* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   NULL,			/* fixups */
   NULL				/* logger */
};

